import { Injectable } from '@angular/core';
import { Observable } from "rxjs/Observable";
// import { isString } from '@angular/compiler/src/facade/lang';
import { Scheduler } from "rxjs/Rx";
import { Subject } from "rxjs/Subject";

export interface WebCallback {
    id: string;
    fun: (event: any) => void;
}

interface CallbackFuns {
    [id: string]: (event: any) => void;
}

@Injectable()
export class $WebSocket {
    public normalCloseCode = 1000;
    private reconnectableStatusCodes = [4000];
    private reconnectAttempts = 0;

    private sendQueue = [];

    // Callbacks 
    private onOpenCallbacks: CallbackFuns = {};
    private onMessageCallbacks: CallbackFuns = {};
    private onErrorCallbacks: CallbackFuns = {};
    private onCloseCallbacks: CallbackFuns = {};

    // Constants 
    private readyStateConstants = {
        'CONNECTING': 0,
        'OPEN': 1,
        'CLOSING': 2,
        'CLOSED': 3,
        'RECONNECT_ABORTED': 4
    };

    // The websocket connection
    private socket: WebSocket;

    // Internal state
    private internalConnectionState: number;

    constructor(private url:string, private protocols?:Array<string>, private config?: WebSocketConfig ) {
        var match = new RegExp('wss?:\/\/').test(url);
        if (!match) {
            throw new Error('Invalid url provided');
        }
        this.config = config ||{ initialTimeout: 500, maxTimeout: 300000, reconnectIfNotNormalClose: false, reconnectTimeout: 10000};
    }

    /***************************
     * Connect/reconnect/close functions
     */
    connect() {
        var self = this;
        if (!this.socket || this.socket.readyState !== this.readyStateConstants.OPEN) {
            this.socket = new WebSocket(this.url);

            this.socket.onopen = (ev: Event) => {
                self.onOpenHandler(ev);
            };

	    this.socket.onmessage = (ev: MessageEvent) => {
                self.onMessageHandler(ev);
            };

            this.socket.onclose = (ev: CloseEvent) => {
                self.onCloseHandler(ev);
            };

            this.socket.onerror = (ev: ErrorEvent) => {
                self.onErrorHandler(ev);
            };
        }
    }

    reconnect() {
        this.close(true);
        console.log('Reconnecting in ' + this.config.reconnectTimeout + ' ms.');
        setTimeout(() => this.connect(), this.config.reconnectTimeout);
    }

    close(force: boolean) {
        if (force || !this.socket.bufferedAmount) {
            this.socket.close();
        }
    }

    /***********************
     * send websocket data.
     */
    send(data) {
        var self = this;
	return new Promise((resolve, reject) => {
	    if (self.socket.readyState === self.readyStateConstants.OPEN) {
		self.sendQueue.push({message: data});
		self.fireQueue();
	    } else {
		reject('Socket connection has been closed.');
	    }
	});
    }
    
    fireQueue() {
        while (this.sendQueue.length && this.socket.readyState === this.readyStateConstants.OPEN) {
            var data = this.sendQueue.shift();
	    this.socket.send(
                (typeof data.message === "string") ? data.message : JSON.stringify(data.message)
            );
        }
    }

    /**************************
     * Add callback functions
     */
    onOpen(cb) {
        this.onOpenCallbacks[cb.id] = cb.fun;
    }

    onClose(cb) {
        this.onCloseCallbacks[cb.id] = cb.fun;
    }

    onError(cb) {
        this.onErrorCallbacks[cb.id] = cb.fun;
    }

    onMessage(cb) {
        this.onMessageCallbacks[cb.id] = cb.fun;
    }

    /**************************
     * Remove callback functions
     */
    removeOnClose(key) {
	delete this.onCloseCallbacks[key];
    }
    
    removeOnError(key) {
	delete this.onErrorCallbacks[key];
    }
    
    removeOnOpen(key) {
	delete this.onOpenCallbacks[key];
    }
    
    removeOnMessage(key) {
	delete this.onMessageCallbacks[key];
    }


    /**************************
     * The callbacks are handled here!
     */
    onMessageHandler(message: MessageEvent) {
        for (var key in this.onMessageCallbacks) {
	    this.onMessageCallbacks[key].apply(this, [message]);
        }
    }    

    onCloseHandler(event: CloseEvent) {
        this.notifyCloseCallbacks(event);
        if ((this.config.reconnectIfNotNormalClose && event.code !== this.normalCloseCode) || this.reconnectableStatusCodes.indexOf(event.code) > -1) {
            this.reconnect();
        }
    }
    
    onErrorHandler(event) {
        this.notifyErrorCallbacks(event);
    }

    onOpenHandler(event: Event) {
        this.reconnectAttempts = 0;
        for (var key in this.onOpenCallbacks) {
	    this.onOpenCallbacks[key](event);
        }
        this.fireQueue();
    }

    notifyCloseCallbacks(event) {
        for (var key in this.onCloseCallbacks) {
            this.onCloseCallbacks[key](event);
        }
    }

    notifyErrorCallbacks(event) {
        for (var key in this.onErrorCallbacks) {
            this.onErrorCallbacks[key](event);
        }
    }

    getReadyState() {
        if (this.socket == null) {
            return -1;
        }
        return this.internalConnectionState || this.socket.readyState;
    }
}

export interface WebSocketConfig {
    initialTimeout: number;
    maxTimeout: number;
    reconnectIfNotNormalClose: boolean;
    reconnectTimeout: number;
}

